#
# (C) Copyright 2020 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
The :class:`~pywbem_mock.InstanceWriteProvider` class provides the default
implementations of the ``CreateInstance``, ``ModifyInstance`` and
``DeleteInstance`` provider methods by means of
:meth:`~pywbem_mock.InstanceWriteProvider.CreateInstance`,
:meth:`~pywbem_mock.InstanceWriteProvider.ModifyInstance`, and
:meth:`~pywbem_mock.InstanceWriteProvider.DeleteInstance`.

A user-defined instance write provider may implement some or all of these
provider methods. The method descriptions linked above provide a detailed
definition of the input parameters, return values, and required behavior.

The following is a simple example of a user-defined instance write provider
that serves the 'CIM_Foo' and 'CIM_FooFoo' classes and implements
``CreateInstance`` for the purpose of setting a value for the 'InstanceID' key
property:

.. code-block:: python

    import uuid
    from pywbem_mock import InstanceWriteProvider

    class MyInstanceProvider(InstanceWriteProvider):

        # CIM classes served by this provider
        provider_classnames = ['CIM_Foo', 'CIM_FooFoo']

        def CreateInstance(self, namespace, new_instance):
            new_instance.properties["InstanceID"] = \\
                f"{new_instance.classname}.{uuid.uuid4()}"
            return super(MyInstanceProvider, self).CreateInstance(
                namespace, new_instance)

Because the provider in this example does not implement the
``ModifyInstance`` and ``DeleteInstance`` methods, these operations will
use their default implementations from
:class:`~pywbem_mock.InstanceWriteProvider`.

NOTE: Bidirectional association support was added in pywbem version 1.5.
originally the association was visible only in the target namespace for the
CreateInstance so Associations and References only handled reference properties
in that namespace.

This instance provider handles association instances that cross namespaces.
Since the association must be available in all of the namespaces defined by the
reference properties of a new instance (References and Associations requests
can be issued using any of the reference properties in the instance), the
CreateInstance creates the association instance in each namespace defined in a
reference property of the instance (it creates shadow instances of the
association in each namespace defined by a reference property in the
association), ModifyInstance modifies the instance and shadow instances, and
DeleteInstance deletes the instance and shadow instances. References and
Association requests can be executed on instances defined in any namespace
where the association is defined.

These shadow instances are visible in each affected namespaces with the
GetInstance and EnumerateInstance requests.

This is transparent. The association instance is visible in each of the
affected namespaces and References and Associations requests can be issued
using any of the reference properties and the operation source instance.

The provider in this example does not need to implement an ``__init__()``
method, because no additional init parameters are needed.
"""


from pywbem._nocasedict import NocaseDict

from pywbem import CIMInstanceName, CIMInstance, CIMError, CIMClass, \
    CIM_ERR_INVALID_PARAMETER, CIM_ERR_ALREADY_EXISTS, CIM_ERR_INVALID_CLASS, \
    CIM_ERR_NOT_FOUND

from pywbem._utils import _format

from ._baseprovider import BaseProvider

# None of the request method names conform since they are camel case
# pylint: disable=invalid-name

__all__ = ['InstanceWriteProvider']


class InstanceWriteProvider(BaseProvider):
    """
    This class defines those instance provider methods that may have
    user-defined providers that override the default provider implementation
    in this class.

    *New in pywbem 1.0 as experimental and finalized in 1.2.*

    User providers are defined by creating a subclass of this class and
    defining a new provider method for one of the methods in this class with
    the same signature.

    Note that user-defined providers may, in turn, call the default providers
    in this class to complete the CreateInstance, ModifyInstance, or
    DeleteInstance, in particular for instance providers that dynamically
    Create or Modify an instance.
    """

    #: str: Keyword defining the type of request the provider will
    #: service. The type for this provider class is predefined as
    #  'instance-write'.
    provider_type = 'instance-write'

    def __init__(self, cimrepository=None):
        """
        Parameters:

          cimrepository (:class:`~pywbem_mock.BaseRepository` or subclass):
            The CIM repository to be used by the provider.
        """
        super().__init__(cimrepository)

    def CreateInstance(self, namespace, new_instance):
        """
        Default provider method for
        :meth:`pywbem.WBEMConnection.CreateInstance`.

        Create a new CIM instance in the CIM repository of the mock WBEM server.

        Validation already performed by the provider dispatcher that calls
        this provider method:

        - The provider method is called only for the registered class and
          namespace (only applies to user-defined providers).

        - The Python types of all input parameters to this provider method are
          as specified below.

        - The namespace exists in the CIM repository.

        - The creation class of the new instance exists in the namespace
          in the CIM repository.

        - All properties specified in the new instance are exposed (i.e.
          defined and inherited with any overrides resolved) by the creation
          class in the CIM repository, and have the same type-related
          attributes (i.e. type, is_array, embedded_object).

        - The creation class of the new instance must not be an abstract class.

        Validation that should be performed by this provider method:

        - new_instance does not specify any properties that are not
          allowed to be initialized by the client, depending on the model
          implemented by the provider.

        - new_instance specifies all key properties needed by the
          provider, depending on the model implemented by the provider.
          The CIM repository will reject any new instance that does not have
          all key properties specified.

        - The new instance (i.e. an instance with the new instance path) does
          not exist in the CIM repository.
          This validation needs to be done by the provider because the
          determination of the key properties for the new instance path may
          depend on the model implemented by the provider.
          The CIM repository will reject the creation of instances that already
          exist, so this check can be delegated to the repository once the
          new instance path has been determined.

        - If new_instance is an association, the reference properties must
          define existing end-point paths to instances or have value
          None.

        Parameters:

          namespace (str):
            The name of the CIM namespace in which the CIM instance is to be
            created, in any lexical case, and with leading and trailing slash
            characters removed.

          new_instance (:class:`~pywbem.CIMInstance`):
            A representation of the CIM instance to be created.

            This object is a deep copy of the original client parameter, and may
            be modified by the provider as needed, before storing it in the
            CIM repository.

            The property names in this object have been adjusted to match the
            lexical case of the property definitions in the creation class
            of the instance in the CIM repository.

            The `classname` attribute of this object will specify the
            creation class for the new instance, in any lexical case.

            The `properties` attribute of this object will specify the
            initial property values for the new CIM instance, with property
            names in any lexical case. Key properties may or may not be
            included.

            The `path` attribute of this object will be `None`.

            The `qualifiers` attribute of this object, if non-empty, should
            be ignored by the provider, because instance-level qualifiers have
            been deprecated in CIM.

        Returns:

            :class:`~pywbem.CIMInstanceName`: Instance path of the new CIM
            instance.

        Raises:
            :exc:`~pywbem.CIMError`: The provider may raise CIMError with any
              status code, and typically raises:
              - CIM_ERR_INVALID_PARAMETER
              - CIM_ERR_ALREADY_EXISTS
        """

        # Get the creation class with all exposed properties and qualifiers.
        # Since the existence of the class has already been verified, this
        # will always succeed.

        class_store = self.cimrepository.get_class_store(namespace)
        creation_class = class_store.get(new_instance.classname, copy=False)

        # Validate association class references have valid end points
        # paths or None
        if self.is_association(creation_class):
            for pn in new_instance:
                prop = new_instance.properties[pn]
                if prop.type == 'reference':
                    if prop.value is None:
                        continue
                    # Exception if end point does not exist
                    self.validate_reference_property_endpoint_exists(prop,)

        # If association class and references multiple namespaces, add instance
        # to each namespace defined in the reference properties
        if self.is_association(creation_class):
            assoc_namespaces = self.find_multins_association_ref_namespaces(
                new_instance, namespace)
            if assoc_namespaces:
                # It is a multi-namespace association instance. Validate
                # characteristics of other namespaces and insert the same
                # instance in each of these namespaces with specific path.
                return self.create_multi_namespace_instance(
                    new_instance, namespace, assoc_namespaces)

        # Otherwise create a single new instance.
        # This default provider determines the instance path from the key
        # properties in the new instance. A user-defined provider may do
        # that as well, or invent key properties such as InstanceID.
        # Specifying strict=True in from_instance() verifies that all key
        # properties exposed by the class are specified in the new instance,
        # and raises ValueError if key properties are missing.
        self.create_new_instance_path(creation_class, new_instance, namespace)
        self.add_new_instance(new_instance)
        return new_instance.path

    def ModifyInstance(self, modified_instance, IncludeQualifiers=None):
        # pylint: disable=invalid-name,line-too-long,unused-argument
        """
        Default provider method for
        :meth:`pywbem.WBEMConnection.CreateInstance`.

        Modify an existing CIM instance in the CIM repository of the mock WBEM
        server.

        NOTE: This method specifies the namespace in modified_instance.path
        rather than as a separate input parameter.

        The modification of the instance in the CIM repository that is performed
        by the provider should be limited to property value changes (including
        addition of properties if not yet set on the instance), because
        instance-level qualifiers have been deprecated in CIM.

        The set of properties that are to be modified in the CIM instance has
        already been determined by the caller so that the modified_instance
        parameter specifies exactly the set of properties to be modified.
        Therefore, this provider method does not have a property list
        parameter.

        Validation already performed by the provider dispatcher that calls
        this provider method:

        - The provider method is called only for the registered class and
          namespace (only applies to user-defined providers).

        - The Python types of all input parameters to this provider method are
          as specified below.

        - The classnames in modified_instance are consistent between
          instance and instance path.

        - The namespace exists in the CIM repository.

        - The creation class of the instance to be modified exists in the
          namespace of the CIM repository.

        - The instance to be modified exists in the namespace of the CIM
          repository.

        - All properties in modified_instance that are to be modified are
          exposed (i.e. defined and inherited with any overrides resolved) by
          the creation class in the CIM repository, and have the same
          type-related attributes (i.e. type, is_array, embedded_object).

        - No key properties are requested to change their values.

        - If there was a property list for the request, only properties defined
          in that property list are included in the modified_instance.

        Validation that should be performed by this provider method:

        - modified_instance does not specify any changed values for
          properties that are not allowed to be changed by the client,
          depending on the model implemented by the provider.

        Validation executed for this default provider:

        - Validate that any reference properties to be modified that are not
          keys define an association end point instance that currently exists
          and that the en-point instances exist.

        - If the association crosses namespaces, validate that the association
          already exists in all namespaces defined in the reference properties.

        Parameters:

          modified_instance (:class:`~pywbem.CIMInstance`):
            A representation of the modified CIM instance, also indicating its
            instance path, with exactly the set of properties to be modified.

            This object is a deep copy of the original client parameter, and may
            be modified by the provider as needed, before storing it in the
            CIM repository.

            The `path` attribute of this object will be set and is the
            instance path of the instance to be modified in the CIM repository.
            Its `namespace`, `classname` and `keybindings` attributes
            will be set. The names will be in any lexical case.

            The `classname` attribute of this object will specify the creation
            class of the instance to be modified, in any lexical case.

            The `properties` attribute of this object will specify exactly the
            set of properties that are to be updated, taking into account the
            original ModifiedInstance and PropertyList input parameters of the
            ModifyInstance() client call.
            The lexical case of the property names has been adjusted to match
            the lexical cae of the property definitions in the creation class
            in the CIM repository.

            The `qualifiers` attribute of this object, if non-empty, should
            be ignored by the provider, because instance-level qualifiers have
            been deprecated in CIM.

          IncludeQualifiers (bool):
            This parameter should be ignored by the provider, because
            instance-level qualifiers have been deprecated in CIM.

        Raises:
            :exc:`~pywbem.CIMError`: The provider may raise CIMError with any
              status code, and typically raises:
              - CIM_ERR_INVALID_PARAMETER
        """  # noqa: E501
        # pylint: disable=invalid-name,line-too-long

        namespace = modified_instance.path.namespace

        # Get a copy of the instance to be modified from the CIM repository.
        instance_store = self.cimrepository.get_instance_store(namespace)
        # Copied because this will be used as the instance to update.
        original_instance = instance_store.get(modified_instance.path)

        # Get class defined for modified instance.
        class_store = self.cimrepository.get_class_store(namespace)
        creation_class = class_store.get(modified_instance.classname,
                                         copy=False)

        # Test for modified reference properties and validate if corresponding
        # instances exist. This is behavior for this default provider.
        # A specific provider may behave differently, ex. ignore non-existend
        # end-point instances
        if self.is_association(creation_class):
            for pn in modified_instance:
                prop = modified_instance.properties[pn]
                if prop.type == 'reference':
                    # Do not allow setting ref property value to None. Would
                    # mean significant extra work removing shadow instances
                    if prop.value is None:
                        raise CIMError(
                            CIM_ERR_INVALID_PARAMETER,
                            _format("Reference property {0!A} association "
                                    "end {1!A} with None value not allowed ",
                                    prop.name, prop.value))
                    if prop.value != original_instance[pn]:
                        self.validate_reference_property_endpoint_exists(prop,)

        # Update the properties in the original instance from properties
        # in the modified instance
        original_instance.update(modified_instance.properties)

        # Note: IncludeQualifiers is completely ignored per DMTF spec.

        # If association class and reference properties define multiple
        # namespaces, modify instance in each namespace defined in the
        # instance.
        if self.is_association(creation_class):
            assoc_namespaces = self.find_multins_association_ref_namespaces(
                original_instance, namespace)
            if assoc_namespaces:
                # It is a multi-namespace association instance. Validate
                # characteristics of other namespaces and insert the same
                # instance in each of these namespaces with specific path.
                self.modify_multi_namespace_instance(
                    original_instance, assoc_namespaces)
                return

        # Replace the instance in the CIM repository with the local copy.
        instance_store.update(original_instance.path, original_instance)

    def DeleteInstance(self, InstanceName):
        """
        Default provider method for
        :meth:`pywbem.WBEMConnection.DeleteInstance`.

        Delete an existing instance in the CIM repository of the mock WBEM
        server.

        NOTE: This method specifies the namespace in InstanceName
        rather than as a separate input parameter.

        The provider is not responsible for determining other instances that
        depend on the instance to be deleted (e.g. association instances
        referencing the instance to be deleted); such dependency detection and
        handling is done elsewhere.

        Validation already performed by the provider dispatcher that calls
        this provider method:

        - The provider method is called only for the registered class and
          namespace (only applies to user-defined providers).

        - The Python types of all input parameters to this provider method are
          as specified below.

        - The namespace exists in the CIM repository.

        - The creation class of the instance to be deleted exists in the
          namespace of the CIM repository.

        - The instance to be deleted exists in the namespace of the CIM
          repository.

        Parameters:

          InstanceName (:class:`~pywbem.CIMInstanceName`):
            The instance path of the instance to be deleted with the
            following attributes:

            * `classname`: Name of the creation class of the instance.
            * `keybindings`: Keybindings of the instance.
            * `namespace`: Name of the CIM namespace containing the instance.
              Will not be `None`.
            * `host`: Will be `None`.
        """

        namespace = InstanceName.namespace

        class_store = self.cimrepository.get_class_store(namespace)
        creation_class = class_store.get(InstanceName.classname, copy=False)

        if not self.is_association(creation_class):
            # Delete the instance from the CIM repository
            instance_store = self.cimrepository.get_instance_store(namespace)
            instance_store.delete(InstanceName)

        else:
            multi_ns = self.find_multins_association_ref_namespaces(
                InstanceName, namespace)
            if multi_ns:
                multi_ns.append(namespace)
                instance_name_copy = InstanceName.copy()
                for ns in multi_ns:
                    instance_name_copy.namespace = ns
                    instance_store = self.cimrepository.get_instance_store(ns)
                    instance_store.delete(instance_name_copy)
            else:
                instance_store = \
                    self.cimrepository.get_instance_store(namespace)
                instance_store.delete(InstanceName)

    def post_register_setup(self, conn):
        """
        Method called by provider registration after registation of provider
        is successful. Using this method is optional for registration cases
        where the provider must execute some activity (ex. modify the
        CIM repository) after successful provider registration).

        Override this method in the user-defined provider subclass to execute
        this method.

        Parameters:

          conn (:class:`~pywbem.WBEMConnection`):
            Current connection which allows client methods to be executed
            from within this method.
        """

    #####################################################################
    #
    # Support methods.  The following methods are support for the
    # Create/Modify/Delete instance default implementations above.
    # They are shown as public because they an be useful in building
    # alternates to the this default Create/Modify/Delete instance
    # implementation.
    #
    #####################################################################
    def find_multins_association_ref_namespaces(self, cim_object,
                                                target_namespace):
        # pylint: disable=line-too-long
        """
        Return list of namespaces in which this association participates
        excluding the target namespace.  If an empty list is returned, this
        is not a multi-namespace association.

        Parameters:

          cim_object (:class:`~pywbem.CIMInstance` or :class:`~pywbem.CIMInstanceName`):  # noqa: E501
            Object from which references/keybindings searched to determine
            if multinamespace object. Must be an association.

          target_namespace (str):
            The namespace name provided with the request call
            (ex. CreateInstance)

        Returns:
            List of namespace names of namespaces in which this
            association instance participates if there are multiple namespaces.
            If no other namespaces are defined by reference properties, it
            returns None.
        """
        # pylint: enable=line-too-long
        if isinstance(cim_object, CIMInstanceName):
            instance_store = self.cimrepository.get_instance_store(
                target_namespace)
            cim_object = instance_store.get(cim_object, copy=False)

        assert isinstance(cim_object, CIMInstance)
        creation_class = self.get_required_class(cim_object,
                                                 target_namespace)
        assert self.is_association(creation_class)

        ref_namespaces = set()
        for inst_prop in cim_object.properties.values():
            if inst_prop.type == 'reference':
                refprop_namespace = inst_prop.value.namespace
                assert refprop_namespace is not None, \
                    _format("Invalid namespace value None found in reference "
                            "property: {0|A}", inst_prop.value)

                # Add to list if namespace exists and not same as
                # target_namespace
                if refprop_namespace:
                    if refprop_namespace != target_namespace:
                        ref_namespaces.add(inst_prop.value.namespace)

        return list(ref_namespaces)

    def get_required_class(self, instance, namespace):
        """
        Get the class defined by instance.classname from the repository. This
        is the common method to validate the class for create instance.

        If the class  does not exist, generate a CIMError.
        """
        class_store = self.cimrepository.get_class_store(namespace)

        try:
            return class_store.get(instance.classname)
        except KeyError:
            raise CIMError(
                CIM_ERR_INVALID_CLASS,
                _format("Creation class {0!A} of new multi-namespace "
                        "association instance does not "
                        "exist in namespace {1!A} of the CIM repository.",
                        instance.classname, namespace))

    @staticmethod
    def is_association(klass):
        """
        Test if klass is an association class.

        Return True if klass is an association class and value is True.
        If association qualifier exists but is False or does not exist return
        False.
        """
        assert isinstance(klass, CIMClass)
        return klass.qualifiers.get('Association', False)

    def create_multi_namespace_instance(self, new_instance, orig_ns,
                                        assoc_namespaces):
        """
        Creates an instance of association new_instance in each namespace
        defined by reference properties of the association instance where the
        namespace in the reference property is not the namespace defined with
        the create instance call.

        Each instance is the same as the original instance except that each
        has its own namespace definition in the CIMInstanceName and the
        reference properties differ if any reference property does not include
        the namespace in its definition.

        Confirm that this is a multi-namespace association instance, that
        the class exists in all namespaces. If these tests are OK,
        confirm that the instance does not exist in any other namespace.
        Then, place the instance in the remote namespaces

        Any case where the instances cannot be created for all required
        namespaces causes an exception.

        Parameters:
          new_instance (:class:`~pywbem.CIMInstance`):
            The new instance that will be created

          orig_ns (str):
            The namespace name attached to the CreateInstance request

          assoc_namespaces (:class:`py:list` of :class:`py:str`):
            List of the namespaces defined in reference properties including
            the original namespace (orig_ns)

        Returns:
          If the instance is created, the CIM_InstanceName for the created
          instance is returned.

        Asserts:
            CIMError if error occurs in validity of other namespaces, class
            in other namespaces, or creating of instance in other namespaces.
        """

        # Add the original namespace to the list.
        assoc_namespaces.append(orig_ns)

        # Validate that the class exists in all of the namespaces.  It was
        # already tested for existence in the namespace defined with the
        # create.
        for ns in assoc_namespaces:
            # Verify that the class exists in the namespace
            creation_class = self.get_required_class(new_instance, ns)

        # Class exists in all target namespaces.  Confirm instances are correct.
        new_instance_paths = {}
        for ns in assoc_namespaces:
            # Set new path in the instance for namespace ns
            # This is the default behavior. A provider may chose to
            # define the path otherwise since it may also be building the
            # instance to be created dynamically.
            # FUTURE: How do we handle that or should we require that this is
            # the only multi-ns method.
            try:
                inst_path = CIMInstanceName.from_instance(
                    creation_class, new_instance, namespace=ns,
                    strict=True)
                new_instance_paths[ns] = inst_path
            except ValueError as exc:
                raise CIMError(CIM_ERR_INVALID_PARAMETER, str(exc))

        # Confirm the new instance does not exist in any of the namespaces.
        for ns, path in new_instance_paths.items():
            instance_store = self.cimrepository.get_instance_store(ns)
            if instance_store.object_exists(path):
                raise CIMError(
                    CIM_ERR_ALREADY_EXISTS,
                    _format("New instance {0!A} already exists in namespace "
                            "{1!A}. Cannot create new instance.", path, ns))

        # Add the new instance with path for each ns to each of the namespaces.
        for ns, path in new_instance_paths.items():
            instance_store = self.cimrepository.get_instance_store(ns)
            new_instance.path = path
            self.add_new_instance(new_instance)

        return new_instance_paths[orig_ns]

    def add_new_instance(self, new_instance):
        """
        Add the new instance to the repository.  If the instance already exists
        raises CIMError exception
        """
        namespace = new_instance.path.namespace
        # Get the instance store of the CIM repository. Since the existence of
        # the namespace has already been verified, this will always succeed.
        instance_store = self.cimrepository.get_instance_store(namespace)

        # Store the new instance in the CIM repository, verifying that it does
        # not exist yet.
        # NOTE: This is where we have invalid repository issue
        try:
            instance_store.create(new_instance.path, new_instance)
        except ValueError:
            raise CIMError(
                CIM_ERR_ALREADY_EXISTS,
                _format("New instance {0!A} already exists in namespace "
                        "{1!A}.", new_instance.path, namespace))

    def modify_multi_namespace_instance(self, modified_instance,
                                        assoc_namespaces):
        """
        Modifies the instance of association modified_instance in each
        namespace defined by reference properties of the association instance
        where the namespace in the reference property is not the namespace
        defined with the create instance call.

        The instances are the same except the instance path contains the
        namespace in which the instance exists.

        Any case where the instances cannot be modified for all required
        namespaces causes an exception.

        Parameters:
          modified_instance (:class:`~pywbem.CIMInstance`):
            The modified instance that is to replace the original
            instance in the repository

          assoc_namespaces (:class:`py:list` of :class:`py:str`):
            List of the namespaces defined in reference properties including
            the original namespace (orig_ns)

        Returns:
          If the instance is created, the CIM_InstanceName for the created
          instance is returned.

        Asserts:
            CIMError if error occurs in validity of other namespaces, class
            in other namespaces, or creating of instance in other namespaces.
        """

        # Add the original namespace as the last in list.
        assoc_namespaces.append(modified_instance.path.namespace)

        # Validate that the class exists in all of the namespaces.  It was
        # already tested for existence in the namespace defined with the
        # create.
        for ns in assoc_namespaces:
            # Verify that the class exists in the namespace
            _ = self.get_required_class(modified_instance, ns)

        # Confirm instances are correct and build dict of instance names for
        # all the assoc intance names
        modified_instance_paths = NocaseDict()  # used to keep dict order
        for ns in assoc_namespaces:
            # Set new path in the instance for namespace ns
            modified_path = modified_instance.copy().path
            modified_path.namespace = ns
            modified_instance_paths[ns] = modified_path

        # Confirm the modified instance exists in all of the namespaces.
        for ns, path in modified_instance_paths.items():
            instance_store = self.cimrepository.get_instance_store(ns)
            if not instance_store.object_exists(path):
                raise CIMError(
                    CIM_ERR_NOT_FOUND,
                    _format("Modified instance {0!A} does not exist in "
                            "namespace {1!A}. Modify Failed", path, ns))

        # Modify the instance path for each namespace
        for ns, path in modified_instance_paths.items():
            instance_store = self.cimrepository.get_instance_store(ns)
            modified_instance.path = path
            instance_store.update(modified_instance.path, modified_instance)

    @staticmethod
    def create_new_instance_path(creation_class, new_instance, namespace):
        """
        Create the path for the new instance and insert it into the
        new_instance.path.

        This default provider determines the instance path from the key
        properties in the new instance. A user-defined provider may do that
        as well, or invent key properties such as InstanceID.
        Specifying strict=True from_instance() verifies that all key
        properties exposed by the class are specified in the new instance,
        and raises ValueError if key properties are missing.

        Parameters:
          creation_class (:class:`~pywbem.CIMClass`):
            The CIM class from which the instance was created. Used to
            determine the key properties.

          new_instance (:class:`~pywbem.CIMInstance`):
            The instance from which the CIMInstanceName will be derived.

          namespace: (str):
            The namespace in which the instance will be added.

        Raises:
            :exc:`~pywbem.CIMError`: if instance name cannot be created from
              instance.

        """
        try:
            new_instance.path = CIMInstanceName.from_instance(
                creation_class, new_instance, namespace=namespace,
                strict=True)

        except ValueError as exc:
            raise CIMError(CIM_ERR_INVALID_PARAMETER, str(exc))

    def validate_instance_exists(self, path):
        """
        Confirm that the path defined by the path exists in the namespace
        defined for that path. Null value for path allowed (i.e. no path
        defined). Returns True if the object exists or the path is empty
        otherwise False.

        Parameters:
            path (:class:`~pywbem.CIMInstanceName`):
                path to instance to be tested including namespace.

        Returns:
            True if path exists in instance store or path is Null.
            False if path exists and is not in instance store.
        """
        if not path:
            return True

        try:
            inst_store = self.cimrepository.get_instance_store(path.namespace)
        except KeyError as ke:
            raise CIMError(CIM_ERR_INVALID_PARAMETER, str(ke))

        return inst_store.object_exists(path)

    def validate_reference_property_endpoint_exists(self, prop):
        """
        Validate that the path defined for the reference property prop
        exists.  Generates INVALID_PARAMETER if it does not exist.

        Parameters:
          prop (:class:`~pywbem.CIMInstanceName`):
            Reference property containing CIMInstanceName as value or
            Null value.

        Raises:
            :exc:`~pywbem.CIMError`:  If value of property is not a valid
              instance or Null.
        """
        path = prop.value

        if path.host:
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Reference property {0!A} association "
                        "end point {1!A} includes host element {2!A}. "
                        "Pywbem_mock does not allow host element in "
                        "reference properties of association instances",
                        prop.name, path, path.host))

        if not self.validate_instance_exists(path):
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Reference property {0!A} association "
                        "end point {1!A} does not exist ",
                        prop.name, path))



"""
nocasedict - A case-insensitive ordered dictionary for Python
"""

from __future__ import absolute_import

from ._version import __version__  # noqa: F401
from ._nocasedict import *  # noqa: F403,F401
from ._hashable import *  # noqa: F403,F401
from ._keyableby import *  # noqa: F403,F401



#
# (C) Copyright 2018 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
A CIM provider creates WBEM server responses to operations defined in DSP0200.

The BaseProvider module contains methods required by both builtin
providers and user-defined providers. This module includes methods for the
following functionality:

  * Managing namespaces in the CIM repository
  * Methods that provide access to specific objects in the CIM repository
    including the processing consistent with filtering the returned objects.
    For example, `get_class(...)` the internal equilavent of the GetClass.

  BaseProvider does not have direct access to the common methods
  defined in _wbemconnection_mock
"""


from pywbem._vendor.nocaselist import NocaseList
from pywbem import CIMError, \
    CIM_ERR_NOT_FOUND, CIMInstance, CIMClass, \
    CIM_ERR_ALREADY_EXISTS, CIM_ERR_INVALID_NAMESPACE, \
    CIM_ERR_NAMESPACE_NOT_EMPTY, \
    WBEMServer
from pywbem._nocasedict import NocaseDict
from pywbem._utils import _format

# pywbem_mock implementation configuration variables that are used in
# request responsders.

__all__ = ['BaseProvider']


class BaseProvider:
    """
    BaseProvider is the top level class in the provider hierarchy and includes
    methods required by both builtin providers and user-defined providers.
    This class is not intended to be executed directly.
    """

    def __init__(self, cimrepository):
        """
        Parameters:

          cimrepository (:class:`~pywbem_mock.BaseRepository` or subclass):
            Defines the repository to be used by providers.  The
            repository is fully initialized but contains only objects defined
            by the :class:`~pywbem_mock.FakedWBEMConnection` methods that
            create objects in the repository
            (ex. :meth:`~pywbem_mock.FakedWBEMConnection.compile_mof_file`)
        """
        self._cimrepository = cimrepository
        self._interop_namespace_names = \
            NocaseList(WBEMServer.INTEROP_NAMESPACES)

    def __repr__(self):
        return _format(
            "{s.__class__.__name__}("
            "provider_classnames={s.provider_classnames!A}, "
            "_interop_namespace_names={s._interop_namespace_names!A}, "
            "_cim_repository={s._cimrepository}",
            s=self)

    @property
    def cimrepository(self):
        """
        :class:`~pywbem_mock.BaseRepository` or subclass: Instance of class
        that represents the CIM repository.

        The CIM repository is the data store for CIM classes, CIM
        instances, and CIM qualifier declarations.  All access to the mocker
        CIM data must pass through this variable to the CIM repository.
        See :class:`~pywbem_mock.BaseRepository` for a description of
        the repository interface.

        The mocker may use subclasses of :class:`~pywbem_mock.BaseRepository`
        for specific functionality. Thus,
        :class:`~pywbem_mock.InMemoryRepository`
        implements an InMemory CIM repository that must be initialized for
        each :class:`~pywbem_mock.FakedWBEMConnection` construction.
        """
        return self._cimrepository

    @property
    def namespaces(self):
        """
        :term:`NocaseList` of :class:`py:str`:
        The names of the namespaces that exist in the CIM repository.
        """
        ns_list = self.cimrepository.namespaces
        assert isinstance(ns_list, NocaseList)
        return ns_list

    @property
    def interop_namespace_names(self):
        """
        :term:`NocaseList` of :class:`py:str`:
        The valid Interop namespace names.

        Only these names may be the Interop namespace and only one
        Interop namespace may exist in a WBEM server environment.
        This list is defined in :attr:`pywbem.WBEMServer.INTEROP_NAMESPACES`.
        """
        ns_list = self._interop_namespace_names
        assert isinstance(ns_list, NocaseList)
        return ns_list

    ################################################################
    #
    #   Methods to manage namespaces
    #
    ################################################################

    def validate_namespace(self, namespace):
        """
        Validates that a namespace is defined in the CIM repository.
        Returns only if namespace is valid. Otherwise it generates an
        exception.

        Parameters:

          namespace (str):
            The name of the CIM namespace in the CIM repository (case
            insensitive). Must not be `None`. Any leading or trailing
            slash characters are ignored.

        Raises:

            :exc:`~pywbem.CIMError`: (CIM_ERR_INVALID_NAMESPACE)
              Namespace does not exist.
        """
        try:
            self.cimrepository.validate_namespace(namespace)
        except KeyError:
            raise CIMError(
                CIM_ERR_INVALID_NAMESPACE,
                _format("Namespace does not exist in CIM repository: {0!A}",
                        namespace))

    def add_namespace(self, namespace, verbose=False):
        """
        Add a CIM namespace to the CIM repository.

        The namespace must not yet exist in the CIM repository and the
        repository can contain only one Interop namespace.

        An Interop namespace can be added with this method, if an Interop
        namespace does not yet exist in the CIM repository.

        Parameters:

          namespace (str):
            The name of the CIM namespace in the CIM repository. Must not be
            `None`. Any leading or trailing slash characters are removed before
            the string is used to define the namespace name.

          verbose (bool):
            Verbose mode: Print a message about the namespace creation.

        Raises:

          ValueError: Namespace argument must not be None.
          :exc:`~pywbem.CIMError`: CIM_ERR_ALREADY_EXISTS if the namespace
            already exists in the CIM repository.
          :exc:`~pywbem.CIMError`: CIM_ERR_ALREADY_EXISTS if the namespace
            is one of the possible Interop namespace names and an interop
            namespace already exists in the CIM repository.
        """

        if namespace is None:
            raise ValueError("Namespace argument must not be None")

        namespace = namespace.strip('/')  # Just for appearance in exc messages

        # Cannot add more than one of the possible Interop namespace names
        if self.is_interop_namespace(namespace):
            if self.find_interop_namespace():
                raise CIMError(
                    CIM_ERR_ALREADY_EXISTS,
                    _format("An Interop namespace {0!A} already exists in the "
                            "CIM repository. {1!A} cannot be added. ",
                            self.find_interop_namespace(), namespace))
        if verbose:
            print(f"Creating namespace {namespace} (in mock support)")
        try:
            self.cimrepository.add_namespace(namespace)
        except ValueError:
            raise CIMError(
                CIM_ERR_ALREADY_EXISTS,
                _format("Namespace {0!A} already exists in the CIM repository ",
                        namespace))

    def remove_namespace(self, namespace, verbose=False):
        """
        Remove a CIM namespace from the CIM repository.

        The namespace must exist in the CIM repository and must be empty.

        The Interop namespace cannot be removed from the CIM repository.

        Parameters:

          namespace (str):
            The name of the CIM namespace in the CIM repository (case
            insensitive). Must not be `None`. Leading or trailing
            slash characters are ignored.

          verbose (bool):
            Verbose mode: Print a message about the namespace deletion.

        Raises:

          ValueError: Namespace argument must not be None.
          :exc:`~pywbem.CIMError`:  CIM_ERR_NOT_FOUND if the namespace does
            not exist in the CIM repository.
          :exc:`~pywbem.CIMError`:  CIM_ERR_INVALID_NAMESPACE if the Interop
            namespace was specified.
          :exc:`~pywbem.CIMError`:  CIM_ERR_NAMESPACE_NOT_EMPTY if the
            namespace icontains objects.
          :exc:`~pywbem.CIMError`:  CIM_ERR_NAMESPACE_NOT_EMPTY if attempting
            to delete the default connection namespace.  This namespace cannot
            be deleted from the CIM repository
        """
        if namespace is None:
            raise ValueError("Namespace argument must not be None")

        namespace = namespace.strip('/')

        if namespace not in self.cimrepository.namespaces:
            raise CIMError(
                CIM_ERR_NOT_FOUND,
                _format("Namespace {0!A} does not exist in the CIM repository ",
                        namespace))

        if self.is_interop_namespace(namespace):
            raise CIMError(
                CIM_ERR_INVALID_NAMESPACE,
                _format("The Interop namespace {0!A} cannot be removed from "
                        "the CIM repository.", namespace))

        if verbose:
            print(f"Deleting namespace {namespace} (in mock support)")
        try:
            self.cimrepository.remove_namespace(namespace)
        # KeyError cannot happen because existence was already verified
        except ValueError:
            raise CIMError(
                CIM_ERR_NAMESPACE_NOT_EMPTY,
                _format("Namespace {0!A} contains objects.", namespace))

    def is_interop_namespace(self, namespace):  # pylint: disable=no-self-use
        """
        Tests if a namespace name is a valid Interop namespace name.

        This method does not access the CIM repository for this test; it
        merely compares the specified namespace name against the list of valid
        Interop namespace names returned by :attr:`interop_namespace_names`.

        Parameters:

          namespace (str):
            The namespace name that is to be tested.

        Returns:

          bool: Indicates whether the namespace name is a valid
          Interop namespace name.
        """
        return namespace in self.interop_namespace_names

    def find_interop_namespace(self):
        """
        Find the Interop namespace name in the CIM repository, or return `None`.

        The Interop namespace is identified by comparing all namespace names
        in the CIM repository against the list of valid Interop namespace names
        returned by :attr:`interop_namespace_names`.

        Returns:

          str: The name of the Interop namespace if one exists in
          the CIM repository or `None`.
        """
        for ns_name in self.cimrepository.namespaces:
            if self.is_interop_namespace(ns_name):
                return ns_name
        return None

    ################################################################
    #
    #   Common Repository access methods used by MainProvider and
    #   InstanceProviders
    #
    ################################################################

    @staticmethod
    def _remove_qualifiers(obj):
        """
        Remove all qualifiers from the input object where the object may
        be an CIMInstance or CIMClass. Removes qualifiers from the object and
        from properties, methods, and parameters

        This is used to process the IncludeQualifier parameter for CIMClass
        and CIMInstance objects.

        Parameters:

          obj(:class:`~pywbem.CIMClass` or :class:`~pywbem.Instance`)
            Object from which qualifiers are to be removed
        """
        assert isinstance(obj, (CIMInstance, CIMClass))
        obj.qualifiers = NocaseDict()
        for prop in obj.properties:
            obj.properties[prop].qualifiers = NocaseDict()
        if isinstance(obj, CIMClass):
            for method in obj.methods:
                obj.methods[method].qualifiers = NocaseDict()
                for param in obj.methods[method].parameters:
                    obj.methods[method].parameters[param].qualifiers = \
                        NocaseDict()

    @staticmethod
    def _remove_classorigin(obj):
        """
        Remove all ClassOrigin attributes from the input object. The object
        may be a CIMInstance or CIMClass.

        Used to process the IncludeClassOrigin parameter of CIMInstance and
        CIMClass objects

        Parameters:

          obj(:class:`~pywbem.CIMClass` or :class:`~pywbem.Instance`)
            Object from which classorigin attribute is to be removed.
        """
        assert isinstance(obj, (CIMInstance, CIMClass))
        for prop in obj.properties:
            obj.properties[prop].class_origin = None
        if isinstance(obj, CIMClass):
            for method in obj.methods:
                obj.methods[method].class_origin = None

    def get_class(self, namespace, classname, local_only=None,
                  include_qualifiers=None, include_classorigin=None,
                  property_list=None):
        # pylint: disable=invalid-name
        """
        Get class from CIM repository.  Gets the class defined by classname
        from the CIM repository, creates a copy, expands the copied class to
        include superclass properties if not localonly, and filters the
        class based on propertylist and includeClassOrigin.

        This method executes all of the filter actions on the class that are
        defined for the GetClass operation and so returns a class that
        satisfies the behavior requirements of the GetClass client request
        operation defined in :term:`DSP0200` .
        (see: :meth:`pywbem.WBEMConnection.GetClass`)

        It also sets the propagated attribute.

        Parameters:

          classname (str):
            Name of class to retrieve

          namespace (str):
            The name of the CIM namespace in the CIM repository (case
            insensitive). Must not be `None`. Leading or trailing slash
            characters are ignored.

          local_only (bool):
            If `True` or `None`, only properties and methods in this specific
            class are returned. `None` means not supplied by client and the
            normal server default is `True`.
            If `False`, properties and methods from the superclasses
            are included.

          include_qualifiers (bool):
            If `True` or `None`, include qualifiers.  `None` is the server
            default if the parameter is not provided by the client.
            If `False`, do not include qualifiers.

          include_classorigin (bool):
            If `True`, return the class_origin attributes of properties and
            methods.
            If `False` or `None` (use server default), class_origin attributes
            of properties and methods are not returned

          property_list (:class:`py:list` of :class:`py:str`):
            Properties to be included in returned class.  If None, all
            properties are returned.  If empty list, no properties are returned

        Returns:

            :class:`~pywbem.CIMClass`: Copy of the CIM class in the CIM
            repository if found. Includes superclass properties installed and
            filtered in accord with the local_only, etc. arguments.

        Raises:

            :exc:`~pywbem.CIMError`: (CIM_ERR_NOT_FOUND) if class not found in
              CIM repository.
            :exc:`~pywbem.CIMError`: (CIM_ERR_INVALID_NAMESPACE) if namespace
              does not exist.
        """

        class_store = self.cimrepository.get_class_store(namespace)

        # Try to get the target class and create a copy for response
        try:
            klass = class_store.get(classname, copy=True)
        except KeyError:
            raise CIMError(
                CIM_ERR_NOT_FOUND,
                _format("Class {0!A} not found in namespace {1!A}.",
                        classname, namespace))

        # local_only server default is True so True or None remove properties
        if local_only is True or local_only is None:
            # Deleting items while iterating requires using list() on py3:
            for pname, prop in list(klass.properties.items()):
                if prop.propagated:
                    del klass.properties[pname]
            for mname, meth in list(klass.methods.items()):
                if meth.propagated:
                    del klass.methods[mname]

        self.filter_properties(klass, property_list)

        # Remove qualifiers if specified.  Note that the server default
        # is to include_qualifiers if include_qualifiers is None
        if include_qualifiers is False:
            self._remove_qualifiers(klass)

        # class_origin default is False so None or False cause removal
        if not include_classorigin:
            self._remove_classorigin(klass)
        return klass

    def class_exists(self, namespace, classname):
        """
        Test if class defined by classname parameter exists in
        CIM repository defined by namespace parameter.

        Returns:

          bool: `True` if class exists and `False` if it does not exist.

        Raises:

          KeyError: Namespace does not exist in the CIM repository
        """
        class_store = self.cimrepository.get_class_store(namespace)
        return class_store.object_exists(classname)

    def _get_instances(self, classname, namespace):
        """
        Get instances of classname from the CIMRepository.  It does not
        get instances of subclasses.

        Parameters:

          classname (str):
            Name of class for which instances will be retrieved.

          namespace (str):
            Repository namespace to search.

        Returns:
          List of instances in repository that represent classname
        """
        classnames = NocaseList(classname)
        instance_store = self.cimrepository.get_instance_store(namespace)

        insts = [self._get_bare_instance(inst.path, instance_store)
                 for inst in instance_store.iter_values()
                 if inst.path.classname in classnames]
        return insts

    @staticmethod
    def filter_properties(obj, property_list):
        """
        Remove properties from an instance or class that aren't in the
        property_list parameter.

        Parameters:

          obj (:class:`~pywbem.CIMClass` or :class:`~pywbem.CIMInstance`):
            The class or instance from which properties are to be filtered

          property_list (:class:`py:list` of :class:`py:str`):
            List of properties which are to be included in the result. If
            None, remove nothing.  If empty list, remove everything. else
            remove properties that are not in property_list. Duplicated names
            are allowed in the list and ignored.
        """
        if property_list is not None:
            property_list = [p.lower() for p in property_list]
            # Deleting items while iterating requires using list() on py3:
            for pname in list(obj.properties.keys()):
                if pname.lower() not in property_list:
                    del obj.properties[pname]

    @staticmethod
    def _get_bare_instance(instance_name, instance_store, copy=None):
        """
        Find an instance in the CIM repository by `instance_name` and return
        that instance. the `copy` parameter controls whether the original
        instance in the CIM repository is returned or a copy.  The
        only time the original should be returned is when the user is
        certain that the returned object WILL NOT be modified.

        DO not modify an instance returned with this method

        Parameters:

          instance_name (:class:`~pywbem.CIMInstanceName`):
            Instance path of the instance to find in the CIM repository

          instance_store (:class:`~pywbem_mock.BaseObjectStore`):
            Instance store of the CIM repository to search for the instance

          copy (bool):
            If True do copy of the instance and return the copy. Otherwise
            return the instance in the CIM repository

        Returns:
            :class:`~pywbem.CIMInstance`: the complete CIM instance or copy of
            the CIM instance is returned if it is found in the CIM repository
            or `None` if the instance defined by `instance_name` is not found
            in the CIM repository. No filtering of properties, qualifiers,
            etc. is executed on the instance.
        """

        if instance_store.object_exists(instance_name):
            return instance_store.get(instance_name, copy=copy)

        return None

    def is_subclass(self, klass, superclass, class_store):
        """
        Return boolean indicating whether a class is a (direct or indirect)
        subclass of a superclass, or the same class, in the specified class
        store of the CIM repository (i.e. in the namespace of that class store).

        This test is done by starting at the class and walking its
        superclasses up to the root, so it is not as expensive as determining
        all subclasses of a given class.

        Parameters:

          klass (str): Class name of the class.

          superclass (str): Class name of the superclass.

          class_store (:class:`~pywbem_mock.BaseObjectStore`):
            Class store of the CIM repository to search for the classes.

        Returns:

            bool: Boolean indicating whether the class is a (direct or indirect)
            subclass of the superclass, or the same class.

        Raises:

            KeyError: Class or superclass does not exist in the class store.
        """

        # This check is at the begin in order to catch the case where class and
        # superclass are the ame class and both do not exist:
        klass_cls = class_store.get(klass)

        if klass.lower() == superclass.lower():
            # Found it
            return True

        next_klass = klass_cls.superclass
        if next_klass is None:
            # Exhausted, we are at the root class
            # Check superclass existence
            class_store.get(superclass)
            return False

        # Continue searching upwards
        return self.is_subclass(next_klass, superclass, class_store)



#
# (C) Copyright 2003-2007 Hewlett-Packard Development Company, L.P.
# (C) Copyright 2006-2007 Novell, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Original authors, until 2007:
#   Tim Potter <tpot@hp.com>
#   Martin Pool <mbp@hp.com>
#   Bart Whiteley <bwhiteley@suse.de>
#
# Authors since 2014:
#   Andreas Maier <andreas.r.maier@gmx.de>
#   Karl Schopmeyer <k.schopmeyer@swbell.net>
#

"""
The only class exposed by this package is :class:`nocasedict.NocaseDict`.
"""

from __future__ import print_function, absolute_import

import sys
import os
import warnings
from collections import OrderedDict
from collections.abc import MutableMapping, KeysView, ValuesView, ItemsView
from typing import Any, AnyStr, NoReturn, Optional, Iterator, Tuple, Dict

from ._utils import _stacklevel_above_nocasedict

__all__ = ['NocaseDict']

# Starting with Python 3.7, the standard dict is guaranteed to be ordered.
# Note: In CPython, that already happened in 3.6, but it was not guaranteed
# for all implementations.
# pylint: disable=invalid-name
if sys.version_info[0:2] >= (3, 7):
    _ODICT_TYPE = dict
else:
    _ODICT_TYPE = OrderedDict

Key = Optional[AnyStr]

# This env var is set when building the docs. It causes the methods
# that are supposed to exist only in a particular Python version, not to be
# removed, so they appear in the docs.
BUILDING_DOCS = os.environ.get('BUILDING_DOCS', False)

# Used as default value for parameters to detect that they have not been
# specified as an argument. Idea from CPython's datetime.timezone.
_OMITTED = object()


class _DictView:
    # pylint: disable=too-few-public-methods
    """
    Base class for directory views, with common methods.
    """

    def __init__(self, dct):
        self._dict = dct

    def __len__(self):
        return len(self._dict)

    def __contains__(self, x):
        return x in iter(self)

    def __reversed__(self):
        return reversed(list(iter(self)))

    def __repr__(self):
        return f"{self.__class__.__name__}({self._dict!r})"


class dict_keys(_DictView, KeysView):
    # pylint: disable=too-few-public-methods
    """
    Dictionary values view.
    """

    def __iter__(self):
        # pylint: disable=protected-access
        data = self._dict._data
        for k in data:
            yield data[k][0]


class dict_values(_DictView, ValuesView):
    # pylint: disable=too-few-public-methods
    """
    Dictionary values view.
    """

    def __iter__(self):
        # pylint: disable=protected-access
        data = self._dict._data
        for k in data:
            yield data[k][1]


class dict_items(_DictView, ItemsView):
    # pylint: disable=too-few-public-methods
    """
    Dictionary items view.
    """

    def __iter__(self):
        # pylint: disable=protected-access
        data = self._dict._data
        for k in data:
            yield data[k]


class NocaseDict(MutableMapping):
    """
    A case-insensitive and case-preserving ordered dictionary.

    The dictionary is case-insensitive: When items of the dictionary are
    looked up by key or keys are compared by the dictionary, that is done
    case-insensitively. The case-insensitivity is defined by performing the
    lookup or comparison on the result of the :meth:`__casefold__` method on
    the key value.
    `None` is allowed as a key value and will not be case folded.

    The dictionary is case-preserving: When keys are returned, they have
    the lexical case that was originally specified when adding or updating the
    item.

    The dictionary is ordered: The dictionary maintains the order in which items
    were added for all Python versions supported by this package. This is
    consistent with the ordering behavior of the built-in :class:`py:dict`
    class starting with Python 3.7.

    The :class:`~nocasedict.NocaseDict` class is derived from the abstract base
    class :class:`py:collections.abc.MutableMapping` and not from the
    :class:`py:dict` class, because of the unique implementation of
    :class:`~nocasedict.NocaseDict`, which maintains a single dictionary with
    the casefolded keys, and values that are tuples (original key, value).
    This supports key based lookup with a single dictionary lookup.
    Users that need to test whether an object is a dictionary should do that
    with ``isinstance(obj, Mapping)`` or ``isinstance(obj, MutableMapping)``.

    The provided key and value objects will be referenced from the
    dictionary without being copied, consistent with the built-in
    :class:`py:dict` class.

    Except for the case-insensitivity of its keys, the
    :class:`~nocasedict.NocaseDict` class behaves like the built-in
    :class:`py:dict` class starting with Python 3.7 (where it is guaranteed to
    be ordered), so its documentation applies completely.

    The :class:`~nocasedict.NocaseDict` class itself provides no added
    functionality compared to the built-in :class:`py:dict` class.
    This package provides mixin classes for adding functionality:

    * :class:`~nocasedict.HashableMixin` mixin class: Adds case-insensitive
      hashability.

    * :func:`~nocasedict.KeyableByMixin` mixin generator function: Adds ability
      to get the key from an attribute of the value object.

    Example of usage::

        from nocasedict import NocaseDict

        dict1 = NocaseDict({'Alpha': 1, 'Beta': 2})

        print(dict1['ALPHA'])  # Lookup by key is case-insensitive
        # 1

        print(dict1)  # Access of keys is case-preserving
        # NocaseDict({'Alpha': 1, 'Beta': 2})
    """

    # Methods not implemented:
    #
    # * __getattribute__(self, name): The method inherited from object is used;
    #   no reason to have a different implementation.
    #
    # * __sizeof__(self): The method inherited from object is used.
    #   TODO(issue #37): Clarify the rules for implementing __sizeof__().

    def __init__(self, *args, **kwargs) -> None:
        """
        Parameters:

          *args : An optional single positional argument representing key-value
            pairs to initialize the dictionary from, in iteration order of the
            specified object. The argument must be one of:

            - a dictionary object, or more specifically an object that has a
              method ``keys()`` providing iteration through the keys and that
              supports subscription by key (e.g. ``ncd[key]``) for accessing
              the values.

            - an iterable. If a key occurs more than once (case-insensitively),
              the last item for that key becomes the corresponding item in
              the dictionary. Each item in the iterable must be one of:

              * an iterable with exactly two items. The first item is used as
                the key, and the second item as the value.

              * an object with a key attribute, if the
                :func:`~nocasedict.KeyableByMixin` mixin generator function is
                used. The value of the key attribute is used as the key, and
                the object itself as the value.

          **kwargs : Optional keyword arguments representing key-value pairs to
            add to the dictionary after being initialized from the positional
            argument.

            If a key being added is already present (case-insensitively) from
            the positional argument, key and value will be updated from the
            keyword argument.

            Before Python 3.7, the order of keyword arguments as specified in
            the call to the method was not guaranteed to be preserved for the
            method implementation, so passing more than one keyword argument
            may have resulted in arbitrary order of items in the dictionary.

        To summarize, only the following types of init arguments are guaranteed
        to preserve the order of provided items after having been added to the
        new dictionary, across all Python versions supported by this package:

        * Passing an iterable as a single positional argument, and passing at
          most one keyword argument.

        * Passing an ordered dictionary/mapping as a single positional
          argument, and passing at most one keyword argument.

        A :exc:`py:UserWarning` will be issued if the order of provided items
        in the arguments is not guaranteed to be preserved.

        Examples for initializing::

            from nocasedict import NocaseDict

            dict1 = NocaseDict({'Alpha': 1, 'Beta': 2})
            dict2 = NocaseDict(dict1)
            dict3 = NocaseDict([('Alpha', 1), ('Beta', 2)])
            dict4 = NocaseDict((('Alpha', 1), ('Beta', 2)))
            dict5 = NocaseDict(Alpha=1, Beta=2)
            dict6 = NocaseDict(dict1, BETA=3)

        Raises:
          TypeError: Expected at most 1 positional argument, got {n}.
          ValueError: Cannot unpack positional argument item #{i}.
        """

        # The internal dictionary, with casefolded keys. An item in this dict
        # is the tuple (original key, value).
        self._data: Dict[Key, Any] = _ODICT_TYPE()

        self.update(*args, **kwargs)

    def _casefolded_key(self, key: Key) -> Key:
        """
        This method returns the casefolded key and handles the case of key
        being `None`.
        """
        if key is None:
            return None
        return self.__casefold__(key)

    @staticmethod
    def __casefold__(key: AnyStr) -> AnyStr:
        """
        This method implements the case-insensitive behavior of the class.

        It returns a case-insensitive form of the input key by calling a
        "casefold method" on the key. The input key will not be `None`.

        The casefold method called by this method is :meth:`py:str.casefold`.
        If that method does not exist on the key value (e.g. because it is a
        byte string), :meth:`py:bytes.lower` is called, for compatibility with
        earlier versions of the package.

        This method can be overridden by users in order to change the
        case-insensitive behavior of the class.
        See :ref:`Overriding the default casefold method` for details.

        Parameters:
            key (AnyStr): Input key. Will not be `None`.

        Returns:
            AnyStr: Case-insensitive form of the input key.

        Raises:
          AttributeError: The key does not have the casefold method.
        """
        try:
            return key.casefold()  # type: ignore
        except AttributeError:
            # Probably a byte string, fall back to lower()
            return key.lower()

    # Basic accessor and setter methods

    def __getitem__(self, key: Key) -> Any:
        """
        Return the value of the item with an existing key (looked up
        case-insensitively).

        Invoked when using e.g.: ``value = ncd[key]``

        Raises:
          AttributeError: The key does not have the casefold method.
          KeyError: Key does not exist (case-insensitively).
        """
        k = self._casefolded_key(key)
        try:
            return self._data[k][1]
        except KeyError:
            key_error = KeyError(f"Key {key!r} not found")
            key_error.__cause__ = None  # Suppress 'During handling..'
            raise key_error  # pylint: disable=raise-missing-from

    def __setitem__(self, key: Key, value: Any) -> None:
        """
        Update the value of the item with an existing key (looked up
        case-insensitively), or if an item with the key does not exist, add an
        item with the specified key and value.

        Invoked when using e.g.: ``ncd[key] = value``

        Raises:
          AttributeError: The key does not have the casefold method.
        """
        k = self._casefolded_key(key)
        self._data[k] = (key, value)

    def __delitem__(self, key: Key) -> None:
        """
        Delete the item with an existing key (looked up case-insensitively).

        Invoked when using: ``del ncd[key]``

        Raises:
          AttributeError: The key does not have the casefold method.
          KeyError: Key does not exist (case-insensitively).
        """
        k = self._casefolded_key(key)
        try:
            del self._data[k]
        except KeyError:
            key_error = KeyError(f"Key {key!r} not found")
            key_error.__cause__ = None  # Suppress 'During handling..'
            raise key_error  # pylint: disable=raise-missing-from

    def __len__(self) -> int:
        """
        Return the number of items in the dictionary.

        Invoked when using: ``len(ncd)``
        """
        return len(self._data)

    def __contains__(self, key: Any) -> bool:
        """
        Return a boolean indicating whether the dictionary contains an item
        with the key (looked up case-insensitively).

        Invoked when using: ``key in ncd``

        Raises:
          AttributeError: The key does not have the casefold method.
        """
        k = self._casefolded_key(key)
        return k in self._data

    def __reversed__(self) -> Iterator[Any]:
        """
        Return an iterator for the reversed iteration order of the dictionary.

        Invoked when using: ``reversed[ncd]``
        """
        # Implementing __reversed__() is necessary because the fall back of
        # reversed() to using len() and __getitem__() (the sequence protocol")
        # requires that the object is a sequence, and relying on the fallback
        # for dicts results in TypeError.
        return reversed(self.keys())

    @classmethod
    def fromkeys(cls, iterable, value=None) -> 'NocaseDict':
        """
        Return a new :class:`NocaseDict` object with keys from the specified
        iterable of keys, and values all set to the specified value.

        Raises:
          AttributeError: The key does not have the casefold method.
        """
        return cls([(key, value) for key in iterable])

    def get(self, key: Key, default=None) -> Any:
        """
        Return the value of the item with an existing key (looked up
        case-insensitively), or if the key does not exist, a default value.

        Raises:
          AttributeError: The key does not have the casefold method.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def pop(self, key: Key, default=_OMITTED) -> Any:
        """
        Remove the item with the specified key if it exists (looked up
        case-insensitively), and return its value.

        If an item with the key does not exist, the default value is returned
        if specified, otherwise :exc:`py:KeyError` is raised.

        Raises:
          KeyError: Key does not exist (case-insensitively) and no default was
            specified.
        """
        k = self._casefolded_key(key)
        try:
            return self._data.pop(k)[1]
        except KeyError:
            if default is not _OMITTED:
                return default
            raise

    def popitem(self) -> Tuple[Key, Any]:
        """
        Remove the last dictionary item (in iteration order) and return it as a
        tuple (key, value).

        The last item in iteration order is the last item that was added to the
        dictionary.

        Raises:
          KeyError: Dictionary is empty.
        """
        return self._data.popitem()[1]

    def setdefault(self, key: Key, default=None) -> Any:
        """
        If an item with the key (looked up case-insensitively) does not exist,
        add an item with that key and the specified default value, and return
        the value of the item with the key.

        Raises:
          AttributeError: The key does not have the casefold method.
        """
        if key not in self:
            self[key] = default
        return self[key]

    # Iteration methods

    def keys(self) -> dict_keys:
        # pylint: disable=line-too-long
        """
        Return a view on the dictionary keys (in the original lexical case) in
        dictionary iteration order.

        See
        `Dictionary View Objects <https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return dict_keys(self)

    def keys_nocase(self) -> KeysView:
        # pylint: disable=line-too-long
        """
        Return a view on the casefolded dictionary keys in dictionary iteration
        order.

        See
        `Dictionary View Objects <https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return self._data.keys()

    def values(self) -> dict_values:
        # pylint: disable=line-too-long
        """
        Return a view on the dictionary values in dictionary iteration order.

        See
        `Dictionary View Objects <https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return dict_values(self)

    def items(self) -> dict_items:
        # pylint: disable=line-too-long
        """
        Return a view on the dictionary items in dictionary iteration order,
        where each item is a tuple of its key (in the original lexical case)
        and its value.

        See
        `Dictionary View Objects on <https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return dict_items(self)

    def __iter__(self) -> Iterator[Key]:
        """
        Return an iterator through the dictionary keys (in the original lexical
        case) in dictionary iteration order.

        Invoked when using: ``for key in ncd``
        """
        for k in self._data:
            yield self._data[k][0]

    # Other stuff

    def __repr__(self) -> str:
        """
        Return a string representation of the dictionary that is suitable for
        debugging.

        The order of items is in dictionary iteration order, and the keys are
        in the original lexical case.

        Invoked when using e.g.: ``repr(ncd)``
        """
        items = [f"{key!r}: {value!r}" for key, value in self.items()]
        items_str = ', '.join(items)
        return f"{self.__class__.__name__}({{{items_str}}})"

    def update(self, *args, **kwargs) -> None:
        # pylint: disable=arguments-differ,signature-differs
        # Note: The signature in Python 3 is: update(self, other=(), /, **kwds)
        #       Since the / marker cannot be used in Python 2, the *args
        #       approach has the same effect, i.e. to ensure that the
        #       parameter can only be specified as a keyword argument.
        """
        Update the dictionary from key/value pairs.

        If a key is already present in the dictionary (looked up
        case-insensitively), its key and value is updated (without affecting
        its position in the dictionary iteration order). Otherwise, an item
        with the key and value is added to the dictionary.

        The provided key and value objects will be referenced from the
        dictionary without being copied, consistent with the built-in
        :class:`py:dict` class.

        Parameters:

          *args : An optional single positional argument representing key-value
            pairs to update the dictionary from, in iteration order of the
            specified object. The argument must be one of:

            - a dictionary object, or more specifically an object that has a
              method ``keys()`` providing iteration through the keys and that
              supports subscription by key for accessing the values.

            - an iterable. If a key occurs more than once (case-insensitively),
              the last item for that key becomes the corresponding item in
              the dictionary. Each item in the iterable must be one of:

              * an iterable with exactly two items. The first item is used as
                the key, and the second item as the value.

              * an object with a key attribute, if the
                :func:`~nocasedict.KeyableByMixin` mixin generator function is
                used. The value of the key attribute is used as the key, and
                the object itself as the value.

          **kwargs : Optional keyword arguments representing key-value pairs to
            update the dictionary from, after having processed the positional
            argument.

            Before Python 3.7, the order of keyword arguments as specified in
            the call to the method was not guaranteed to be preserved for the
            method implementation, so passing more than one keyword argument
            may have resulted in arbitrary order of items in the dictionary.

        Raises:
          AttributeError: The key does not have the casefold method.
          TypeError: Expected at most 1 positional argument, got {n}.
          ValueError: Cannot unpack positional argument item #{i}.
        """
        if args:
            if len(args) > 1:
                raise TypeError(
                    "Expected at most 1 positional argument, got {n}".
                    format(n=len(args)))
            other = args[0]
            try:
                # Try mapping / dictionary
                for key in other.keys():
                    self[key] = other[key]
                # pylint: disable=unidiomatic-typecheck
                if type(other) is dict and _ODICT_TYPE is not dict and \
                        len(other.keys()) > 1:
                    warnings.warn(
                        "Before Python 3.7, initializing or updating a "
                        "NocaseDict object from a dict object with more than "
                        "one item is not guaranteed to preserve the order of "
                        "its items",
                        UserWarning,
                        stacklevel=_stacklevel_above_nocasedict())
            except AttributeError:
                # Expecting an iterable

                # Try whether KeyableByMixin() was used
                key_attr = getattr(
                    self, 'nocasedict_KeyableByMixin_key_attr', None)
                # The following raises TypeError if not iterable:
                for i, item in enumerate(other):
                    if key_attr and hasattr(item, key_attr):
                        # Is a keyable object
                        key = getattr(item, key_attr)
                        value = item
                    else:
                        # Expecting key, value pair
                        try:
                            key, value = item
                        except ValueError as exc:
                            value_error = ValueError(
                                "Cannot unpack positional argument item #{i} "
                                "of type {t} into key, value: {exc}".
                                format(i=i, t=type(item), exc=exc))
                            value_error.__cause__ = None  # Suppress 'During..'
                            # pylint: disable=raise-missing-from
                            raise value_error
                    self[key] = value

        for key, val in kwargs.items():
            self[key] = val
        if len(kwargs) > 1 and _ODICT_TYPE is not dict:
            warnings.warn(
                "Before Python 3.7, initializing or updating a NocaseDict "
                "object from more than one keyword argument is not guaranteed "
                "to preserve their order",
                UserWarning,
                stacklevel=_stacklevel_above_nocasedict())

    def clear(self) -> None:
        """
        Remove all items from the dictionary.
        """
        self._data.clear()

    def copy(self) -> 'NocaseDict':
        """
        Return a copy of the dictionary.

        This is a middle-deep copy; the copy is independent of the original in
        all attributes that have mutable types except for:

        * The values in the dictionary

        Note that the Python functions :func:`py:copy.copy` and
        :func:`py:copy.deepcopy` can be used to create completely shallow or
        completely deep copies of objects of this class.
        """
        result = NocaseDict()
        result._data = self._data.copy()  # pylint: disable=protected-access
        return result

    def __eq__(self, other: Any) -> bool:
        """
        Return a boolean indicating whether the dictionary and the other
        dictionary are equal, by matching items (case-insensitively) based on
        their keys, and then comparing the values of matching items for
        equality.

        The other dictionary may be a :class:`NocaseDict` object or any other
        mapping. In all cases, the matching of keys takes place
        case-insensitively.

        Invoked when using e.g.: ``ncd == other``

        Raises:
          AttributeError: The key does not have the casefold method.
        """
        # Issue #1062: Could compare hash values for better performance
        for key, self_value in self.items():
            if key not in other:
                return False
            other_value = other[key]
            try:
                if not self_value == other_value:
                    return False
            except TypeError:
                return False  # not comparable -> considered not equal
        return len(self) == len(other)

    def __ne__(self, other: Any) -> bool:
        """
        Return a boolean indicating whether the dictionary and the other
        dictionary are not equal, by negating the equality test.

        The other dictionary may be a :class:`NocaseDict` object or any other
        mapping. In all cases, the matching of keys takes place
        case-insensitively.

        Invoked when using e.g.: ``ncd != other``

        Raises:
          AttributeError: The key does not have the casefold method.
        """
        return not self == other

    def _raise_ordering_not_supported(self, other: Any, op: str) -> NoReturn:
        """
        Function to raise a TypeError indicating that ordering of this class
        is not supported.
        """
        raise TypeError(
            "'{}' not supported between instances of '{}' and '{}'".
            format(op, type(self), type(other)))

    def __lt__(self, other: Any) -> NoReturn:
        self._raise_ordering_not_supported(other, '<')

    def __gt__(self, other: Any) -> NoReturn:
        self._raise_ordering_not_supported(other, '>')

    def __ge__(self, other: Any) -> NoReturn:
        self._raise_ordering_not_supported(other, '>=')

    def __le__(self, other: Any) -> NoReturn:
        self._raise_ordering_not_supported(other, '<=')



# Copyright (C) 2020 Andreas Maier
"""
nocaselist - A case-insensitive list for Python
"""

from __future__ import absolute_import

from ._version import __version__  # noqa: F401
from ._nocaselist import *  # noqa: F403,F401



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
The following warnings are pywbem specific warnings that can be issued by
the WBEM client library.
"""

from builtins import Warning as builtin_Warning

from ._exceptions import Error

# This module is meant to be safe for 'import *'.

__all__ = ['Warning', 'ToleratedServerIssueWarning',
           'ToleratedSchemaIssueWarning', 'MissingKeybindingsWarning',
           'OldNameFilterWarning', 'OldNameDestinationWarning',
           'RequestExceptionWarning']


class Warning(Error, builtin_Warning):
    # pylint: disable=redefined-builtin
    """
    Base class for pywbem specific warnings.
    """
    pass


class ToleratedServerIssueWarning(Warning):
    """
    This warning indicates an issue with the WBEM server that has been
    tolerated by pywbem.
    """
    pass


class ToleratedSchemaIssueWarning(Warning):
    """
    This warning indicates that a component in a DMTF CIM Schema is
    invalid but the issue is tolerated or corrected by pywbem.
    """
    pass


class MissingKeybindingsWarning(Warning):
    """
    This warning indicates that an instance path without keybindings has been
    encountered, either when sending a CIM-XML request to a WBEM server or
    when receiving a CIM-XML response from a WBEM server.

    A local creation of :class:`~pywbem.CIMinstanceName` objects without
    keybindings does not issue this warning.
    """
    pass


class OldNameFilterWarning(Warning):
    """
    This warning indicates that an owned indication filter instance with an old
    name format (prior to pywbem 1.3) was discovered on the WBEM server.

    Such filters are ignored when discovering owned filters. They should be
    cleaned up by the user.
    """
    pass


class OldNameDestinationWarning(Warning):
    """
    This warning indicates that an owned listener destination instance with an
    old name format (prior to pywbem 1.3) was discovered on the WBEM server.

    Such destinations are ignored when discovering owned destinations. They
    should be cleaned up by the user.
    """
    pass


class RequestExceptionWarning(Warning):
    """
    This warning indicates that an unexpected behavior occurred with excptions
    raised by the "requests" package.
    """
    pass



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
"""
Pywbem_mock supports several variables that provide configuration and
behavior control of the mock environment.

These configuration variables can be modified by the user directly after
importing pywbem. For example:

::

    import pywbem_mock
    pywbem_mock.config.SYSTEMNAME = 'MyTestSystemName'

Note that the pywbem source file defining these variables should not be changed
by the user. Instead, the technique shown in the example above should be used to
modify the configuration variables.
"""

# This module is meant to be safe for 'import *'.

__all__ = ['OBJECTMANAGERNAME', 'SYSTEMNAME',
           'SYSTEMCREATIONCLASSNAME', 'DEFAULT_MAX_OBJECT_COUNT',
           'OPEN_MAX_TIMEOUT', 'OBJECTMANAGERCREATIONCLASSNAME',
           'IGNORE_INSTANCE_IQ_PARAM', 'IGNORE_INSTANCE_ICO_PARAM']

#: Name for the object manager It is used in defining user-defined provider
#: properties for CIM_Namespace provider and CIM_ObjectManager provider if
#: if they are defined.
OBJECTMANAGERNAME = 'FakeObjectManager'

#: Value for the CIM_ObjectManagerCreationClassName defined in the
#: CIM_ObjectManager class.
OBJECTMANAGERCREATIONCLASSNAME = 'CIM_ObjectManager'

#: The name of the mock object. This string value becomes part of the
#: CIM_ObjectNamager instance and CIM_Namespace instances
SYSTEMNAME = 'MockSystem_WBEMServerTest'

#: Name for the property SystemCreationClassname defined a number of
#: CIM classes that are used by the pywbem_mock including CIM_Namespace
SYSTEMCREATIONCLASSNAME = 'CIM_ComputerSystem'

#: Default value for the MaxObjectCount parameter of the Open...() methods of
#: the mock WBEM server, if the value was not provided by the user.
DEFAULT_MAX_OBJECT_COUNT = 100

#: Maximum value for the OperationTimeout parameter of the Open...() methods of
#: the mock WBEM server.
OPEN_MAX_TIMEOUT = 40

#: Use of the IncludeQualifiers parameter on instance requests is DEPRECATED in
#: DMTF DSP0200. The definition of IncludeQualifiers is ambiguous and when this
#: parameter is set to true, WBEM clients cannot be assured that any qualifiers
#: will be returned. A WBEM client should always set this parameter to false. To
#: minimize the impact of this recommendation on WBEM clients, a WBEM server may
#: choose to treat the value of IncludeQualifiers as false for all requests.
#:
#: The following variable forces pywbem_mock to ignore the client supplied
#: variable and not return qualifiers on EnumerateInstances and GetInstance
#: responses.
#:
#: * True (default): pywbem_mock always removes qualifiers from instances
#:   in responses
#: * False: pywbem_mock uses value of input parameter or its default to
#:   determine if qualifiers are to be removed
IGNORE_INSTANCE_IQ_PARAM = True

#: Use of the IncludeClassOrigin parameter on instance requests is DEPRECATED. A
#: WBEM server may choose to treat the value of IncludeClassOrigin parameter as
#: false for all requests, otherwise the implementation shall support the
#: original behavior as defined in the rest of this paragraph. If the
#: IncludeClassOrigin input parameter is true, the CLASSORIGIN attribute shall
#: be present on all appropriate elements in each returned Instance. If it is
#: false, no CLASSORIGIN attributes are present.
#:
#: The following variable forces pywbem_mock to ignore the client supplied
#: variable and not return qualifiers on EnumerateInstances and GetInstance
#: responses.
#:
#: * True (default): pywbem_mock always removes class origin attributes from
#:   instances in responses
#: * False: pywbem_mock uses value of input parameter or its default to
#:   determine if class origin attributes are to be removed
IGNORE_INSTANCE_ICO_PARAM = True



#
# (C) Copyright 2020 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
This module is part of the support for user-defined providers.  User-defined
providers may be created within pywbem_mock to extend the capability
of the mocker for special processing for selected classes.

This module contains the ProviderDispatcher class which routes request methods
that allow user-defined providers either to the default method  or if a
user-defined provider is registered to the user-defined method for the
provider class defined in the provider registry.
"""


from copy import deepcopy
import warnings
from collections.abc import Mapping, Sequence

from pywbem import CIMInstance, CIMInstanceName, CIMClass, CIMClassName, \
    CIMParameter, CIMError, CIM_ERR_NOT_FOUND, CIM_ERR_INVALID_PARAMETER, \
    CIM_ERR_INVALID_CLASS, CIM_ERR_METHOD_NOT_FOUND, cimtype, \
    ToleratedSchemaIssueWarning
from pywbem._utils import _format
from pywbem._nocasedict import NocaseDict

from ._baseprovider import BaseProvider
from ._instancewriteprovider import InstanceWriteProvider
from ._methodprovider import MethodProvider

__all__ = ['ProviderDispatcher']


class ProviderDispatcher(BaseProvider):
    """
    This class dispatches requests destined for the provider methods defined in
    InstanceWriteProvider or MethodProvider  to the either the default provider
    method or the provider registered for processing the defined class in the
    defined namespace.

    It handles the following requests:

    * CreateInstance: default is InstanceWriteProvider.CreateInstance
    * ModifyInstance: default is InstanceWriteProvider.ModifyInstance
    * DeleteInstance: default is InstanceWriteProvider.DeleteInstance
    * InvokeMethod: default is MethodProvider.InvokeMethod
    """

    def __init__(self, cimrepository, provider_registry):
        """
        Set instance parameters passed from FakedWBEMConnection
        """
        super().__init__(cimrepository)

        self.provider_registry = provider_registry

        # Define the instances of the default implementations for the
        # InstanceWriteProvider and MethodProvider. The default providers
        # are constructed on first call.
        self._default_instance_write_provider = None
        self._default_method_provider = None

    @property
    def default_instance_write_provider(self):
        """
        Instance object for default instance provider. Constructed on first
        call
        """
        if self._default_instance_write_provider is None:
            self._default_instance_write_provider = InstanceWriteProvider(
                self.cimrepository)
        return self._default_instance_write_provider

    @property
    def default_method_provider(self):
        """
        Instance object for default method provider. Constructed on first
        call
        """
        if self._default_method_provider is None:
            self._default_method_provider = MethodProvider(
                self.cimrepository)
        return self._default_method_provider

    def _validate_property(
            self, prop_name, instance, creation_class, namespace, class_store):
        """
        Validate a property of an instance against its declaration in a class.
        """

        if prop_name not in creation_class.properties:
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Property {0!A} in the instance does not exist in its "
                        "creation class {1!A} in namespace {2!A} of the CIM "
                        "repository",
                        prop_name, instance.classname, namespace))

        prop_inst = instance.properties[prop_name]
        prop_cls = creation_class.properties[prop_name]

        if prop_inst.type != prop_cls.type:
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Property {0!A} in the instance has incorrect "
                        "type={1!A}, but should have type={2!A} according to "
                        "its creation class {3!A} in namespace {4!A} of the "
                        "CIM repository",
                        prop_name, prop_inst.type, prop_cls.type,
                        instance.classname, namespace))

        if prop_inst.is_array != prop_cls.is_array:
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Property {0!A} in the instance has incorrect "
                        "is_array={1!A}, but should have is_array={2!A} "
                        "according to its creation class {3!A} in namespace "
                        "{4!A} of the CIM repository",
                        prop_name, prop_inst.is_array, prop_cls.is_array,
                        instance.classname, namespace))

        if isinstance(prop_inst.value, CIMInstance):
            emb_classname_inst = prop_inst.value.classname
            if 'EmbeddedInstance' in prop_cls.qualifiers:
                ei_qual = prop_cls.qualifiers['EmbeddedInstance']
                emb_classname_cls = ei_qual.value
                if not self.is_subclass(
                        emb_classname_inst, emb_classname_cls, class_store):
                    raise CIMError(
                        CIM_ERR_INVALID_PARAMETER,
                        _format("Property {0!A} in the instance is an embedded "
                                "instance of class {1!A}, but should be of "
                                "class {2!A} according to its creation class "
                                "{3!A} in namespace {4!A} of the CIM "
                                "repository",
                                prop_name, emb_classname_inst,
                                emb_classname_cls,
                                instance.classname, namespace))
            elif 'EmbeddedObject' not in prop_cls.qualifiers:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Property {0!A} in the instance is an embedded "
                            "instance of class {1!A}, but the property "
                            "declaration has neither EmbeddedInstance nor "
                            "EmbeddedObject set in its creation class {2!A} in "
                            "namespace {3!A} of the CIM repository",
                            prop_name, emb_classname_inst,
                            instance.classname, namespace))

        if isinstance(prop_inst.value, CIMClass):
            emb_classname_inst = prop_inst.value.classname
            if 'EmbeddedObject' not in prop_cls.qualifiers:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Property {0!A} in the instance is an embedded "
                            "class {1!A}, but the property declaration does "
                            "not have EmbeddedObject set in its creation class "
                            "{2!A} in namespace {3!A} of the CIM repository",
                            prop_name, emb_classname_inst,
                            instance.classname, namespace))

    def CreateInstance(self, namespace, NewInstance):
        # pylint: disable=invalid-name
        """
        Dispatcher for the CreateInstance provider method.

        This method performs validations and if successful, routes the provider
        method call either to a registered provider, or to the default provider.
        """

        # Verify the input parameter types (type errors have already been
        # raised during checks in the WBEMConnection operation).
        assert isinstance(namespace, str)
        assert isinstance(NewInstance, CIMInstance)

        # Verify that the new instance has no path (ensured by the
        # WBEMConnection operation).
        assert NewInstance.path is None

        # Verify that the namespace exists in the CIM repository.
        self.validate_namespace(namespace)

        # Get creation class from CIM repository. The CIMClass objects in the
        # class store of the repository have all exposed properties (i.e.
        # defined and inherited, having resolved all overrides), qualifiers,
        # and classorigjn information.
        class_store = self.cimrepository.get_class_store(namespace)
        try:
            creation_class = class_store.get(NewInstance.classname)
        except KeyError:
            raise CIMError(
                CIM_ERR_INVALID_CLASS,
                _format("Creation class {0!A} of new instance does not "
                        "exist in namespace {1!A} of the CIM repository.",
                        NewInstance.classname, namespace))

        if creation_class.qualifiers.get("Abstract", False):
            warnings.warn(
                _format("Tolerating instance creation of abstract "
                        "class {0} in namepace {1} which is forbidden "
                        "by DMTF DSP0004.",
                        NewInstance.classname, namespace),
                ToleratedSchemaIssueWarning, 1)

        # Verify that the properties in the new instance are exposed by the
        # creation class and have the correct type-related attributes.
        for pn in NewInstance.properties:
            self._validate_property(
                pn, NewInstance, creation_class, namespace, class_store)

        # The providers are guaranteed to get a deep copy of the original
        # new instance since they may update properties.
        new_instance = deepcopy(NewInstance)

        # Adjust the lexical case of the property names in the new instance
        # to match the lexical case of the property definitions in the creation
        # class.
        for inst_pn in new_instance.properties:
            inst_prop = new_instance.properties[inst_pn]
            cls_pn = creation_class.properties[inst_pn].name
            if inst_pn != cls_pn:
                inst_prop.name = cls_pn  # modifies new_instance

        # Determine the provider to be used. Note that a registered provider
        # always has all provider methods for the provider type, either
        # implemented or inherited.
        provider = self.provider_registry.get_registered_provider(
            namespace, 'instance-write', new_instance.classname)
        if not provider:
            provider = self.default_instance_write_provider

        # Call the provider method.
        result = provider.CreateInstance(namespace, new_instance)

        # Verify provider method result.
        assert isinstance(result, CIMInstanceName)

        return result

    def ModifyInstance(self, ModifiedInstance,
                       IncludeQualifiers=None, PropertyList=None):
        # pylint: disable=invalid-name
        """
        Dispatcher for the ModifyInstance provider method.

        This method performs validations and if successful, routes the provider
        method call either to a registered provider, or to the default provider.
        """

        # Verify the input parameter types (type errors have already been
        # raised during checks in the WBEMConnection operation).
        assert isinstance(ModifiedInstance, CIMInstance)
        assert isinstance(IncludeQualifiers, (bool, type(None)))
        assert isinstance(PropertyList, (str, list, tuple, type(None)))
        assert isinstance(ModifiedInstance.path, CIMInstanceName)

        # Verify equality of the class names in the modified instance.
        if ModifiedInstance.classname.lower() != \
                ModifiedInstance.path.classname.lower():
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Modified instance has inconsistent class names: "
                        "{0!A} in the instance, and {1!A} in the instance "
                        "path.",
                        ModifiedInstance.classname,
                        ModifiedInstance.path.classname))

        # Verify that the namespace exists in the CIM repository.
        namespace = ModifiedInstance.path.namespace
        self.validate_namespace(namespace)

        class_store = self.cimrepository.get_class_store(namespace)
        instance_store = self.cimrepository.get_instance_store(namespace)

        # Get creation class from CIM repository and verify that it exists.
        # The CIMClass objects in the class store of the repository have all
        # exposed properties (i.e. defined and inherited, having resolved all
        # overrides), qualifiers, and classorigjn information.
        try:
            creation_class = class_store.get(ModifiedInstance.classname)
        except KeyError:
            raise CIMError(
                CIM_ERR_INVALID_CLASS,
                _format("Creation class {0!A} of modified instance does not "
                        "exist in namespace {1!A} of the CIM repository.",
                        ModifiedInstance.classname, namespace))

        # Get instance to be modified from CIM repository.
        try:
            instance = instance_store.get(ModifiedInstance.path)
        except KeyError:
            raise CIMError(
                CIM_ERR_NOT_FOUND,
                _format("Instance to be modified does not exist in the CIM "
                        "repository: {0!A}",
                        ModifiedInstance.path))

        # Verify that the properties in the property list are exposed by the
        # creation class, and reduce property list to be unique.
        if PropertyList is None:
            property_list = None
        else:
            property_list = []
            property_dict = NocaseDict()
            for pn in PropertyList:
                if pn not in creation_class.properties:
                    raise CIMError(
                        CIM_ERR_INVALID_PARAMETER,
                        _format("Property {0!A} in PropertyList does not "
                                "exist in creation class {1!A} in namespace "
                                "{2!A} of the CIM repository",
                                pn, ModifiedInstance.classname, namespace))
                if pn not in property_dict:
                    property_dict[pn] = True
                    property_list.append(pn)

        # Verify that the properties in the modified instance are exposed by the
        # creation class and have the correct type-related attributes.
        # Strictly, we would only need to verify the properties to be modified
        # as reduced by the PropertyList.

        for pn in ModifiedInstance.properties:

            self._validate_property(
                pn, ModifiedInstance, creation_class, namespace, class_store)

            prop_inst = ModifiedInstance.properties[pn]
            prop_cls = creation_class.properties[pn]

            if prop_cls.qualifiers.get('key', False) and \
                    prop_inst.value != instance[pn]:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Property {0!A} in the instance is a key "
                            "property and thus cannot be modified, "
                            "according to its creation class {1!A} in "
                            "namespace {2!A} of the CIM repository",
                            pn, ModifiedInstance.classname, namespace))

        # The providers are guaranteed to get a deep copy of the original
        # modified instance since they may update properties.
        modified_instance = deepcopy(ModifiedInstance)

        # Reduce modified_instance to have just the properties to be modified
        if property_list is not None:

            # Add class default values for properties not specified in
            # ModifiedInstance.
            for pn in property_list:
                if pn not in modified_instance:
                    # If the property in the class does not have a default
                    # value, it is None.
                    modified_instance[pn] = creation_class.properties[pn].value

            # Remove properties from modified_instance that are not in
            # PropertyList.
            for pn in list(modified_instance):
                if pn not in property_dict:
                    del modified_instance[pn]

        # Adjust the lexical case of the properties in the modified instance to
        # the lexical case they have in the creation class.
        for pn in modified_instance.properties:
            inst_prop = modified_instance.properties[pn]
            cl_prop = creation_class.properties[pn]
            if inst_prop.name != cl_prop.name:
                inst_prop.name = cl_prop.name  # changes modified_instance

        # Determine the provider to be used. Note that a registered provider
        # always has all provider methods for the provider type, either
        # implemented or inherited.
        provider = self.provider_registry.get_registered_provider(
            namespace, 'instance-write', modified_instance.classname)
        if not provider:
            provider = self.default_instance_write_provider

        # Call the provider method.
        result = provider.ModifyInstance(
            modified_instance, IncludeQualifiers=IncludeQualifiers)

        # Verify provider method result.
        assert result is None

    def DeleteInstance(self, InstanceName):
        # pylint: disable=invalid-name
        """
        Dispatcher for the DeleteInstance provider method.

        This method performs validations and if successful, routes the provider
        method call either to a registered provider, or to the default provider.
        """

        # Verify the input parameter types (type errors have already been
        # raised during checks in the WBEMConnection operation).
        assert isinstance(InstanceName, CIMInstanceName)

        # Verify that the namespace exists in the CIM repository.
        namespace = InstanceName.namespace
        self.validate_namespace(namespace)

        class_store = self.cimrepository.get_class_store(namespace)
        instance_store = self.cimrepository.get_instance_store(namespace)

        # Verify that the creation class of the instance to be deleted exists
        # in the CIM repository.
        if not class_store.object_exists(InstanceName.classname):
            raise CIMError(
                CIM_ERR_INVALID_CLASS,
                _format("Creation class {0!A} of instance to be deleted does "
                        "not exist in namespace {1!A} of the CIM repository.",
                        InstanceName.classname, namespace))

        # Verify that the instance to be deleted exists in the CIM repository.
        if not instance_store.object_exists(InstanceName):
            raise CIMError(
                CIM_ERR_NOT_FOUND,
                _format("Instance to be deleted does not exist in the CIM "
                        "repository: {0!A}",
                        InstanceName))

        # Determine the provider to be used. Note that a registered provider
        # always has all provider methods for the provider type, either
        # implemented or inherited.
        provider = self.provider_registry.get_registered_provider(
            namespace, 'instance-write', InstanceName.classname)
        if not provider:
            provider = self.default_instance_write_provider

        # Call the provider method.
        result = provider.DeleteInstance(InstanceName)

        # Verify provider method result.
        assert result is None

    def InvokeMethod(self, methodname, localobject, params):
        # pylint: disable=invalid-name
        """
        Dispatcher for the InvokeMethod provider method.

        This method performs validations and if successful, routes the provider
        method call either to a registered provider, or to the default provider.

        Parameters:

          methodname (str): Method name

          localobject (CIMInstanceName or CIMClassName): Target object, with
            namespace set. Types are validated.

          params (NocaseDict): Input parameters, as follows:
            * key (str): Parameter name.
            * value (CIMParameter): Parameter value.
            Types are validated.

        Returns:

          A tuple of (returnvalue, outparams), with these tuple items:

            * returnvalue (CIM data type): Return value.

            * outparams (NocaseDict): Output parameters, with:
              * key (str): Parameter name
              * value (CIM data type): Parameter value
        """

        namespace = localobject.namespace

        # Verify the input parameter types (type errors have already been
        # raised during checks in WBEMConnection.InvokeMethod(), and in
        # FakedWBEMConnection._mock_methodcall()).
        assert isinstance(namespace, str)
        assert isinstance(methodname, str)
        assert isinstance(localobject, (CIMInstanceName, CIMClassName))
        assert isinstance(params, NocaseDict)

        # Verify that the namespace exists in the CIM repository.
        self.validate_namespace(namespace)

        class_store = self.cimrepository.get_class_store(namespace)
        instance_store = self.cimrepository.get_instance_store(namespace)

        if isinstance(localobject, CIMInstanceName):
            # instance-level use

            # Get the creation class of the target instance from the CIM
            # repository, verifying that it exists.
            try:
                klass = class_store.get(localobject.classname)
            except KeyError:
                raise CIMError(
                    CIM_ERR_INVALID_CLASS,
                    _format("Creation class {0!A} of target instance does "
                            "not exist in namespace {1!A} of the CIM "
                            "repository.",
                            localobject.classname, namespace))

            # Verify that the target instance exists in the CIM repository.
            if not instance_store.object_exists(localobject):
                raise CIMError(
                    CIM_ERR_NOT_FOUND,
                    _format("Target instance does not exist in the CIM "
                            "repository: {0!A}",
                            localobject))

        else:
            assert isinstance(localobject, CIMClassName)
            # class-level use

            # Get the target class from the CIM repository, verifying that it
            # exists.
            try:
                klass = class_store.get(localobject.classname)
            except KeyError:
                raise CIMError(
                    CIM_ERR_NOT_FOUND,
                    _format("Target class {0!A} does not exist in namespace "
                            "{1!A} of the CIM repository.",
                            localobject.classname, namespace))

        # Verify that the class exposes the CIM method.
        if methodname not in klass.methods:
            raise CIMError(
                CIM_ERR_METHOD_NOT_FOUND,
                _format("Method {0!A} is not exposed by class {1!A} in "
                        "namespace {2!A} of the CIM repository.",
                        methodname, klass.classname, namespace))
        method = klass.methods[methodname]

        if isinstance(localobject, CIMClassName):
            # class-level use

            # Verify that the method is static.
            # Note: A similar check for instance-level use is not appropriate
            # because static methods can be invoked on instances or on classes.
            static_qual = method.qualifiers.get('Static')
            static_value = static_qual.value if static_qual else False
            if not static_value:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Non-static method {0!A} in class {1!A} in "
                            "namespace {2!A} cannot be invoked on a class "
                            "object.",
                            methodname, klass.classname, namespace))

        # Verify that the input parameters are defined by the method and have
        # the correct type-related attributes.
        for pn in params:
            assert isinstance(pn, str)
            param_in = params[pn]
            assert isinstance(param_in, CIMParameter)

            if pn not in method.parameters:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("The specified input parameter {0!A} is not "
                            "defined in method {1!A} of class {2!A} in "
                            "namespace {3!A} of the CIM repository",
                            pn, methodname, klass.classname, namespace))

            param_cls = method.parameters[pn]

            in_qual = method.qualifiers.get('In')
            in_value = in_qual.value if in_qual else True
            if not in_value:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("The specified input parameter {0!A} is "
                            "defined as an output-only parameter according to "
                            "its method {1!A} of class {2!A} in namespace "
                            "{3!A} of the CIM repository",
                            pn, methodname, klass.classname, namespace))

            if param_in.type != param_cls.type:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("The specified input parameter {0!A} has "
                            "incorrect type={1!A}, but should have type={2!A} "
                            "according to its method {3!A} in class {4!A} in "
                            "namespace {5!A} of the CIM repository",
                            pn, param_in.type, param_cls.type,
                            methodname, klass.classname, namespace))

            if param_in.is_array != param_cls.is_array:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("The specified input parameter {0!A} has "
                            "incorrect is_array={1!A}, but should have "
                            "is_array={2!A} "
                            "according to its method {3!A} in class {4!A} in "
                            "namespace {5!A} of the CIM repository",
                            pn, param_in.is_array, param_cls.is_array,
                            methodname, klass.classname, namespace))

            if param_in.embedded_object != param_cls.embedded_object:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("The specified input parameter {0!A} has "
                            "incorrect embedded_object={1!A}, but should have "
                            "embedded_object={2!A} "
                            "according to its method {3!A} in class {4!A} in "
                            "namespace {5!A} of the CIM repository",
                            pn, param_in.embedded_object,
                            param_cls.embedded_object,
                            methodname, klass.classname, namespace))

        # Determine the provider to be used.
        provider = self.provider_registry.get_registered_provider(
            namespace, 'method', klass.classname)
        if not provider:
            provider = self.default_method_provider

        # Call the provider method
        result = provider.InvokeMethod(methodname, localobject, params)

        # Verify provider method result
        if not isinstance(result, (list, tuple)):
            raise TypeError(
                _format("InvokeMethod provider method returned invalid type: "
                        "{0}. Must return list/tuple (return value, output "
                        "parameters)",
                        type(result)))
        if len(result) != 2:
            raise ValueError(
                _format("InvokeMethod provider method returned invalid number "
                        "of items: {0}. Must be list/tuple (return value, "
                        "output parameters)",
                        len(result)))
        return_value = result[0]
        output_params = result[1]

        # Map the more flexible way output parameters can be returned from
        # the provider method to what _mock_methodcall() expects
        output_params_dict = NocaseDict()
        if isinstance(output_params, Sequence):
            # sequence of CIMParameter
            for param in output_params:
                if not isinstance(param, CIMParameter):
                    raise TypeError(
                        _format("InvokeMethod provider method returned invalid "
                                "type for item in output parameters "
                                "sequence: {0}. Item type must be "
                                "CIMParameter",
                                type(param)))
                output_params_dict[param.name] = param.value
        elif isinstance(output_params, Mapping):
            # mapping of name:value or name:CIMParameter
            for pname in output_params:
                pvalue = output_params[pname]
                if isinstance(pvalue, CIMParameter):
                    pvalue = pvalue.value
                else:
                    # Perform check for valid CIM data type:
                    try:
                        cimtype(pvalue)
                    except TypeError:
                        new_exc = TypeError(
                            _format("InvokeMethod provider method returned "
                                    "invalid type for value in output "
                                    "parameters mapping: {0}. Value type must "
                                    "be a CIM data type or CIMParameter",
                                    type(pvalue)))
                        new_exc.__cause__ = None
                        raise new_exc
                    except ValueError:
                        # Empty array
                        pass
                output_params_dict[pname] = pvalue
        else:
            raise TypeError(
                _format("InvokeMethod provider method returned invalid type "
                        "for output parameters: {0}. Must be "
                        "Sequence(CIMParameter) or "
                        "Mapping(name: value/CIMParameter)",
                        type(output_params)))

        return return_value, output_params_dict
